function [M, cost, infeasible] = bi_bmatch(W, b, perfect, verbose, useGurobi)
% bibmatch  Solve max-weight (perfect) variable bipartite b-matching
%   [M, cost, infeasible] = bibmatch(W, b, perfect, verbose) takes a
%   "compact weight matrix" W, an NxN matrix where the rows and columns
%   index the two parts of the bipartition.
%
%   W can be dense or sparse, and the sparsity of M depends on W. If W is
%   dense, a zero entry still indicates an edge (of zero cost).
%
%   b is a vector of neighbors. If perfect is true, then we find an exact
%   b-matching. Otherwise, b is an *upper* bound.
%
%   cost stores the weight of the maximum matching, and infeasible is set
%   if the problem has no solution.
%
%   if useGurobi is true, we use gurobi_linprog. Be careful about running
%   multiple processes of Gurobi on one machine. (Default false)
%
%   Currently this implements the LP relaxation, but we could swap out
%   another implementation if it also works.


    % Default options
    if nargin < 5
        useGurobi = false;
    end
            
    [N, NN] = size(W);
    Nsq = N*N;
    assert(N == NN, 'W must be square');
    
    % Massage b
    if length(b) == 1
        % Expand to vector
        bb = b * ones(2*N, 1);        
    elseif isrow(b)
        bb = horzcat(b, b)';
    elseif iscolumn(b)
        bb = vertcat(b, b);
    else
        error('bi_bmatch:b', 'b must be vector or scalar');
    end
    
    % linprog minimizes. Maximize by minimizing the negative.
    f = -W(:);    
    
    % TODO: Smarter way... but this works
    
    % Make it sparse.
    iVec(2*Nsq) = 0;
    jVec(2*Nsq) = 0;    
    ind = 1;
    
    for i = 1:N
        for j = 1: N
            iVec(ind) = i;
            jVec(ind) = N*(i-1) + j;
            
            iVec(ind+1) = i + N;
            jVec(ind+1) = N*(j-1) + i;
            
            ind = ind + 2;
        end
    end    
    
    assert(ind - 1 == 2*Nsq, 'Wrong number of indices!');
    A = sparse(iVec, jVec, ones(2*Nsq, 1), 2*N, Nsq);
    
    Aeq = []; beq = [];
    % Perfect matching: move the the eq constraints.
    if perfect
        Aeq = A;
        beq = bb;
        A = [];
        bb = [];
    end
        
    % TODO: Figure out Gurobi iteration limit.
    if useGurobi
        [x, fval, exitflag] = gurobi_linprog(f, A, bb, Aeq, beq, zeros(size(f)), ones(size(f)));        
    else
        [x, fval, exitflag] =        linprog(f, A, bb, Aeq, beq, zeros(size(f)), ones(size(f)));
    end
    
    if exitflag == -2
        warning('bi_bmatch:infeasible', 'Infeasible');
        M = [];
        cost = [];
        infeasible = true;
        return;
    elseif exitflag ~= 1
        error('bi_bmatch:converge', 'Failed to converge; exit flag was %d', exitflag);
    end
    
    % TODO: Check that the solution is integral
    xr = round(x);
    assertElementsAlmostEqual(xr, x);
    
    M = zeros(N,N);
    for i = 1:N
        for j = 1:N
            M(i,j) = x(N*(j-1)+i);
        end
    end
    
    MM = reshape(x, N, N);
    assert(all(vec(M == MM))); % Figure out what it's doing    
    
    cost = -fval;    
end

