function [P]=edmondsreduction(W,b1)
%
% function [P]=edmondsreduction(W,b1)
%
% Given a weight matrix, runs unipartite Edmonds' algorithm
% to find the maximum weight matching
% via BlossomV from Kolmogorov using reduction to 1-matching
% by replicating nodes and edges b times. See Gerards 1995
% Tutte (1954b), and Bondy&Murty 2008 on page 431.
%
% W is the weight matrix (only the upper triangle matters)
% b1 is the number of edges per node
%
% Example:
% n=6;Q=rand(n,n);for i=1:n;Q(i,1:i)=0;end;[P]=edmondsreduction(Q,2);
% Example in Bondy&Murty
% Q=zeros(6,6);Q(1,2)=1;Q(1,4)=1;Q(1,6)=1;Q(2,3)=1;Q(3,4)=1;Q(3,5)=1;Q(4,5)=1;Q(4,6)=1.1;Q(5,6)=1;
% Solution is P(1,2);P(1,6);P(2,3);P(3,5);P(4,5);P(4,6);
%
%
%     0     1     0     0     0     1
%     1     0     1     0     0     0
%     0     1     0     0     1     0
%     0     0     0     0     1     1
%     0     0     1     1     0     0
%     1     0     0     1     0     0
%
% Qo=rand(n,n);for i=1:n;Qo(i,1:i)=0;end;P2=edmonds2(Qo,2);P=edmondsreduction(Qo,2);[sum(sum(P.*Qo)) sum(sum(P2.*Qo))]
%
% Another test, reverse weights should give same answer
% 
% n=6; b=2;
% Q=rand(n,n);for i=1:n;Q(i,1:i)=0;end;tic;P=edmondsreduction(Q,b); toc;
% nQ=max(max(Q))-Q;for i=1:n;nQ(i,1:i)=0;end;tic;nP=edmondsreduction(nQ,n-b-1);toc;
% nnP=1-nP;for i=1:n;nnP(i,1:i)=0;end;nnP=max(nnP,nnP');imagesc(nnP-P);
%
% If b<n/2, we do the reverse-weight solution since it's faster!
%
n  = size(W,1);
zz = pwd;
if (b1<(n/2))
  b=n-b1-1;
  Wbb=W;
else
  b=b1;
  Wbb=-W;
end

minW = min(Wbb(1,2:n));
maxW = max(Wbb(1,2:n));
for i=2:(n-1)
  minW = min([minW Wbb(i,(i+1):n)]);
  maxW = max([maxW Wbb(i,(i+1):n)]);
end
Wbb = round(20000000*(Wbb-minW)/(maxW-minW)+1);
for i=1:n
  for j=1:(i-1)
    Wbb(i,j) = Wbb(j,i);
  end
end

graphPath = tempname;
outPath   = tempname;

fid = fopen(graphPath,'w');
fprintf(fid,'c graph in DIMACS format \n');
E=zeros(n,n);
ind=1;
for i=1:n
  for j=(i+1):n
    E(i,j)=ind; ind=ind+1;
  end
end
for j=1:n
  for i=(j+1):n
    E(i,j)=ind; ind=ind+1;
  end
end
fprintf(fid,'p edge %d %d\n',n*(n-1)+n*(n-1-b),n*(n-1)/2+n*(n-1-b)*(n-1));

minmin=min(min(Wbb));
for i=1:n
  for j=1:n
    if (i~=j)
      if (E(i,j)<E(j,i))
        fprintf(fid,'e %d %d %d\n',E(i,j),E(j,i),Wbb(i,j));
      end
    end
  end
end
for i=1:n
  for j=1:n
    if (i~=j)
      for k=1:(n-1-b) % Add (degree-b), but degree=(n-1) in full graph
        fprintf(fid,'e %d %d %d\n',E(i,j),n*(n-1)+(i-1)*(n-1-b)+k,0); % (i,j));
      end
    end
  end
end
fclose(fid);
cmd = sprintf('./blossom/blossom5 -V -e %s -w %s', graphPath, outPath);
[status, out] = system(cmd);
fprintf('Status = %d ; output = %s\n', status, out);

Ps=load(outPath);
Ps=Ps+1;
P=zeros(n,n);
m=n*(n-1);
for k=2:size(Ps,1)
  edge1=Ps(k,1);
  edge2=Ps(k,2);
  if (max(edge1,edge2)<=m)
    [i,j]=find(E==edge1);
    [ii,jj]=find(E==edge2);
    P(i,j)=1;
    P(j,i)=1;
    P(ii,jj)=1;
    P(jj,ii)=1;
  end
end

if (b1<(n/2))
  P=1-P;
  for i=1:n
    P(i,1:i)=0;
  end;
  P=max(P,P');
end

cd(zz)


