function [Gfinal, Sfinal, Wfinal] = gvbm(X, kk, runLog, varargin)
% gvbm  General variational b-matching with at least k neighbors
%
%   [Gfinal, Sfinal] = gvbm(X, k, runLog) 
%                    = gvbm(X, k, runLog, 'PropertyName', PropertyValue)
%
%   X is the binary data matrix and k is a scalar or vector of neighbor
%   counts. Prints progress to runLog.
%
%   Output adjacency matrix (between the bipartition, so it may not be
%   symmetric!) in Gfinal and the star pattern in Sfinal.
%
%   Available properties:
%       - Ginit -- initial adjacency matrix. Must be a valid solution.
%       - symmetric (true/false) -- use bipartiter or unipartiter solver
%       - greedy -- whether annealing sticks to the lowest so far or not
%       - negScale, posScale -- annealing schedule; see code
%       - maxIter
%

    p = inputParser;
    p.addRequired('X');
    p.addRequired('k', @isnumeric);
    p.addRequired('runLog', @ischar);
    p.addParamValue('Ginit', []);
    p.addParamValue('Winit', []);
    p.addParamValue('tempInit', -1);
    p.addParamValue('symmetric', false);
    p.addParamValue('greedy', false);
    p.addParamValue('negScale', 2);
    p.addParamValue('posScale', 0.92);
    p.addParamValue('maxIter', 200);
    p.addParamValue('metropolis', false);
    p.addParamValue('roundMe', true);
    p.parse(X, kk, runLog, varargin{:});
    
    o = p.Results;
    
    [N, D] = size(X);

    if isempty(o.Winit)
        % Initialize W to 1... but normalize here.    
        W = ones(N, D);
        % Maybe remove this line for the big datasets? On the other hand,
        % unimatch converges, so might as well keep it.
        W = W / sum(W(:)); 
    else
        W = o.Winit;
    end
    
    Wfinal = W;

    Sold = N*D;    
    if ~isempty(o.Ginit)
        G    = o.Ginit;
        Gold = o.Ginit;
    else        
        Gold = zeros(N,N);
    end

    fid = fopen(runLog, 'a');

    Sfinal = ones(size(X));
    Strail = [];

    maxK = max(kk);
    fprintf(1,  'n=%d d=%d maxk=%d\n', N, D, maxK);
    fprintf(fid,'n=%d d=%d maxk=%d\n', N, D, maxK);

    temperature = o.tempInit; % negative value just means (steal it).

    for iter = 1:o.maxIter
        GW(N,N) = 0;
        
        for i = 1:N
            for j = 1:N
                GW(i,j) = sum(W(i,:) .* (X(i,:)~=X(j,:)));
            end
        end
        
        %GW = GW + rand(size(GW)) * 1e-6;
        
        % Special case: If we start with an initial value, we skip the
        % first iteration
        %
        % Probably a smarter way... but this works!
        if iter > 1 || isempty(o.Ginit)        
            % Drop the excesses        
            if o.symmetric
                GW = triu(GW);
                Gc = double(bmatch(GW, N - kk, false, false, o.roundMe));                
            else
                % CODE TO CORRECT DO IMPERFECT BIPARTITE B-MATCHING... EITHER
                % DO THE LP (Gurobi?) AND DOUBLE-CHECK CONVERGENCE.
                %
                % I think Gurobi crashes when you try to run in two MATLAB
                % processes? Usually in the checkusername routine...
                % Double-check the licensing?

                %Gc = bi_bmatch(GW, N - 1 - kk, false, false, true); % Gurobi needed for accuracy
                Gc = bi_bmatch(GW, N - kk, false, false, true); % Gurobi needed for accuracy
            end
            G = ones(size(Gc)) - Gc;            
        end
                
        % NOTE: This is NOT actually the Hamming error! (It has an
        % extraneous weight term; in the paper, this value is purely
        % discrete.)
        fprintf(1,   'Old Hamming error: %e; ', sum(vec(Gold .* GW)));
        fprintf(fid, 'Old Hamming error: %e; ', sum(vec(Gold .* GW)));
        
        fprintf(1,   'New Hamming error %e\n', sum(vec(G .* GW)));
        fprintf(fid, 'New Hamming error %e\n', sum(vec(G .* GW)));
        
        % Starify.
        S = zeros(size(X));
        [iVec, jVec] = find(G);
        for idx = 1:length(iVec)
            i = iVec(idx);
            j = jVec(idx);

            disagree = X(i,:) ~= X(j,:);
            S(i,disagree) = 1;
        end                        

        Snew = sum(S(:));
        if Snew < sum(Sfinal(:))
            Sfinal = S;
            Gfinal = G;
            Wfinal = W;
        end


        % Metropolis move
        
        if o.metropolis
            Sdiff = Snew - Sold;
            if Sdiff < 0 || rand < exp(-Sdiff / temperature);
                % Accept
                Gold = G;
                Sold = Snew;
            else
                % Reject
                G = Gold;
                Snew = Sold;
            end
        else
            %Don't move to a higher energy state, if we asked for greedy.
            if o.greedy && Snew > Sold
                G = Gold;
                Snew = Sold;
            else
                Gold = G;
                Sold = Snew;
            end                        
        end
        
        for i=1:N
            for k=1:D
                W(i,k) = sum(G(i,:).*(X(i,k)~=X(:,k)'));
            end
        end

        if temperature < 0
            temperature = o.negScale * max(W(:));
        else
            temperature = o.posScale * temperature;
        end

        fprintf(1,  'Iter=%d Stars=%d temperature=%f\n',iter,Snew,temperature);
        fprintf(fid,'Iter=%d Stars=%d temperature=%f\n',iter,Snew,temperature);        

        W=-W/temperature;
        %W=W-max(max(W));  % Why???
        W=exp(W);

        Strail = [Strail Snew];

        % If the last 40 iterations didn't change the star count
        if iter > 40 && length(unique(Strail(end-40:end))) <= 1
            fclose(fid);
            return;
        end

    end

    fclose(fid);    
end
