function [Gmin, Smin] = gvbm_fixed(X, bb, varargin)
% gvbm  General variational b-privacy matching, no anneal.
%
%   [Gfinal, Sfinal] = gvbm(X, k, runLog) 
%                    = gvbm(X, k, runLog, 'PropertyName', PropertyValue)
%
%   X is the binary data matrix and k is a scalar or vector of neighbor
%   counts. Prints progress to runLog.
%
%   Note that b-privacy means each record is indistinguishable from b - 1
%   other records.
%
%   Output adjacency matrix (between the bipartition, so it may not be
%   symmetric!) in Gfinal and the star pattern in Sfinal.
%
%   Available properties:
%       - Ginit -- initial adjacency matrix. Must be a valid solution.
%       - symmetric (true/false) -- use bipartiter or unipartiter solver
%       - greedy -- whether annealing sticks to the lowest so far or not
%       - negScale, posScale -- annealing schedule; see code
%       - maxIter
    
    %% Parse input
    p = inputParser;
    p.addRequired('X');
    p.addRequired('k', @isnumeric);
    p.addParamValue('epsilon', 1e-2);
    p.addParamValue('boundEps', []);
    p.addParamValue('fudgeFactor', []);
    p.addParamValue('annealRate', 1.0);
    p.addParamValue('runLog', 'gbvm_fixed_runlog');
    p.addParamValue('Ginit', []);    
    p.addParamValue('symmetric', false);    
    p.addParamValue('maxIter', 200);
    p.parse(X, bb, varargin{:});
    
    o = p.Results;
    
    [N, D] = size(X);

    %% Initialize
    W = ones(N, D);
    %W = W / sum(W(:));
    Smin   = Inf;
    Gmin   = [];
    Sbound = N*D;    
    SboundOld = Sbound;
    
    if ~isempty(o.Ginit)
        G = o.Ginit;        
    end    
    
    if isempty(o.boundEps)
        o.boundEps = 0.01 * o.epsilon;
    end
    
    if isempty(o.fudgeFactor)
        o.fudgeFactor = 100 * o.boundEps;
    end

    fid = fopen(o.runLog, 'a');

    S = ones(size(X));
    Strail = [];

    maxK = max(bb);
    fprintf(1,  'n=%d d=%d maxk=%d\n', N, D, maxK);
    fprintf(fid,'n=%d d=%d maxk=%d\n', N, D, maxK);

    %% Iterate
    for iter = 1:o.maxIter
        %% Make weighted b-matching problem.
        GW = zeros(N, N);        
        for i = 1:N
            for j = 1:N
                GW(i,j) = sum(W(i,:) .* (X(i,:)~=X(j,:)));
            end
        end
        
        % Break ties. A small amount of noise can't hurt. (Can it?) Yes it
        % can.
        %minGW = min(vec(GW(GW ~= 0)));
        %GW = GW + eps(minGW) * abs(rand(size(GW)));
                
        % Minimize the *complement* problem.
        % See Eq (2) of imperfect_b_matching.pdf.
        if iter > 1 || isempty(o.Ginit)        
            % Drop the excesses        
            if o.symmetric
                GW = triu(GW);
                %G = double(bmatch(GW, bb - 1, false, false));                
                Gc = double(bmatch(GW, N - bb, false, false, false));
            else
                % Note that we don't do kk - 1, because this is a "compact
                % bipartite adjacency matrix", i.e. the diagonal does not
                % mean a self-loop. We actually need the value b.
                %G = bi_bmatch(GW, bb, false, false, true); % Gurobi needed for accuracy
                Gc = bi_bmatch(GW, N - 1 - bb, false, false, true);
            end
            G = ones(size(Gc)) - Gc;
        end
        
        degVec = sum(G, 2);                
        fprintf(1, 'G degrees: min = %d, med = %d, max = %d\n', ...
                min(degVec), median(degVec), max(degVec));                                        
        
        %% Anneal
        o.epsilon = o.annealRate * o.epsilon;
        
        % Paper says \ln \frac{\epsilon}{1 + \epsilon}
        % Implement with log1p            
        logFactor = log(o.epsilon) - log1p(o.epsilon);
        boundLogFactor = log(o.boundEps) - log1p(o.boundEps);
        
        %% Calculate new weights and variational bound (want logNewW to be small)
        logNewW = zeros(size(W));        
        for i=1:N
            for k=1:D
                logNewW(i,k) = sum(G(i,:) .* (X(i,k)~=X(:,k)'));
            end
        end        
        
        %% Compute step by step
        % Scale W for the bound calculation only (do we lose precision?!)
        sW = W / sum(W(:));
        entropy = -sum(vec(sW .* log(sW)));               
        inside   = logNewW * boundLogFactor + log1p(o.boundEps);        
        
        % NOTE: For Sbound to be decreasing, logBound has to be increasing.
        logBound = sum(vec(sW .* inside)) + entropy;        
        
        SboundOld = Sbound;       
        Sbound   = N*D - exp(logBound);
        
        dSbound = Sbound - SboundOld;
        if dSbound > o.fudgeFactor
            warning('gvbm_fixed:mono', 'Monotonicity failed; difference was %g\n', dSbound);
        end
        
        W = exp(logNewW * logFactor);
        %W = W / sum(W(:));
                        
        % What does the err mean?!
        err = sum(vec(G .* GW));
                
        %% Starify.
        S = zeros(size(X));
        [iVec, jVec] = find(G);
        for idx = 1:length(iVec)
            i = iVec(idx);
            j = jVec(idx);

            disagree = X(i,:) ~= X(j,:);
            S(i,disagree) = 1;        
        end                
        Snew = sum(S(:));
        
        Sbdiff = Snew - Sbound;
        if Sbdiff > o.fudgeFactor
            warning('gvbm_fixed:stars', 'Star bound exceeded fudge factor; difference was %g\n', Sbdiff);
        end        
        
        % Store the latest value
        if Snew < Smin
            Smin = Snew;
            Gmin = G;
        end
        
        fprintf(1,  'Iter=%d Stars=%d, Star Bound = %g, Edges = %d, Error=%g, Epsilon = %g, LogFactor = %g\n', iter, Snew, Sbound, sum(G(:)), err, o.epsilon, logFactor);
        fprintf(fid,'Iter=%d Stars=%d, Star Bound = %g, Edges = %d, Error=%g, Epsilon = %g, LogFactor = %g\n', iter, Snew, Sbound, sum(G(:)), err, o.epsilon, logFactor);        
                

        %% Termination condition
        Strail = [Strail Snew];
        % If the last 20 iterations didn't change the star count
        if iter > 40 && length(unique(Strail(end-40:end))) <= 1
            fclose(fid);
            return;
        end
    end

    fclose(fid);    
end
