function [IDX, G, S] = fast_agglomerative(X, k)
% fast_agglomerative  Fast agglomerative clustering, at least k per cluster
%
%   This version explicitly trackers clusters, using logical vectors.
    
    [N, D] = size(X);
    
    template = struct('ones',  false(1, D), ...
        'stars', false(1, D), ...
        'inds',  false(1, N), ...
        'size', 0, ...
        'nStars', 0, ...
        'sizeTimesStars', 0);        
    
    function ss = makeSingletons(cs)
        % Split one cluster cs into a structure array of singletons ss
        
        ns = cs.size;
        ss(ns) = template;
        nzIdxs = find(cs.inds);
        for n = 1:ns
            i = nzIdxs(n);
            ss(n) = template;
            ss(n).ones = X(i,:);
            ss(n).size = 1;
            ss(n).inds(i) = 1;
        end
        
        ss = ss';
        
    end
    
    % Prepare the first iteration
    % Each cluster template can be represented by its ones and stars. (The
    % zeros are the complement of (ones | stars)). Other variables are for
    % memoization.               
    
    % A fake cluster containing everything (all the indices)
    clusters = makeSingletons(struct('inds', true(1, N), 'size', N));
    
    % Dummy singleton; entry 1 will be overwritten. Just to establish the
    % structure.
    finalClusters = makeSingletons(clusters(1));
    nFinalClusters = 0;
        
    iter = 0;    
    while length(clusters) > 1
        dVec = pdist(clusters, @clusterDist);                
                
        [d, r, c] = minPdist(dVec);
        % Need the almost because of the perturbation
        %assertElementsAlmostEqual(d, clusterDist(clusters(r), clusters(c)));
        %clusterDist(clusters(1), clusters(2))
        % It's lower-triangular, so c < r. Merge cluster r into c, and
        % remove r.
        clusters(c) = mergeClusters(clusters(c), clusters(r));        
        clusters(r) = [];
                
        % If it's big enough, remove the cluster
        if clusters(c).size >= k
            % It's >= because k-anonymity means you have k - 1 neighbors.
            nFinalClusters = nFinalClusters + 1;
            finalClusters(nFinalClusters) = clusters(c);
            clusters(c) = [];
        end
        
        % This isn't going to work because your cluster structure is
        % currently greedy-additive... whatever, it works well enough.
%         while clusters(c).size > k
%             % Make a temporary copy that we will do trial deletions from
%             cc = clusters(c);           
%             
%             nCC = length(cc);
%             nzIdxs = find(cc);
%             ddVec(nCC) = 0;
%             for ni = 1:nCC
%                 n = nzIdxs(ni);
%                 % Temporarily remove
%                 cc.inds(n) = 0;
%                 
%                 ddVec(ni) = clusterDist(clusters(c), cc);
%                 
%                 cc.inds(n) = 1;                
%             end
%             
%             [~, niMax] = max(ddVec);
%             nMax = nzIdxs(niMax);
%             
%             % Remove the entry niMax from clusters(c) and put it in its own
%             % singleton cluster
%             clusters(c).inds(nMax) = 0;
%             clusters(c).
%             
%             cs = struct('inds', false(1, N), 'size', 1);
%             cs.inds(nMax) = 1;
%             clusters(end+1) = makeSingleton(cs);
%             
%             
%         end
                            
        iter = iter + 1;
        
        fprintf(1, 'Agglomerative iter = %d; nFinalClusters = %d\n', iter, nFinalClusters);
    end
    
    % Oh, that's the redistribution part that was in Tony's code.
    % Our distance code was written for clusters, so make some singletons
    if length(clusters) == 1        
        ss = makeSingletons(clusters(1));
        nSS = length(ss);
        dMat(nSS, nFinalClusters) = 0;

        for n = 1:nSS
            dMat(n,:) = clusterDist(ss(n), finalClusters');
        end

        [~, iMin] = min(dMat, [], 2);
        for n = 1:nSS
            dst = iMin(n);
            finalClusters(dst) = mergeClusters(finalClusters(dst), ss(n));
        end
    end
                
    % Double check: Each row should match its cluster template. That is, if
    % a dimension is not starred, then all elements within the cluster
    % should have the same value. Note that the *presence* of a star need
    % not imply anything.
    for k = 1:nFinalClusters
        % Zeros are neither stars nor ones
        theZeros = ~(finalClusters(k).ones | finalClusters(k).stars);                        

        for i = find(finalClusters(k).inds)
            onesMatch = finalClusters(k).ones & X(i,:);
            assert(all(finalClusters(k).ones == onesMatch), 'A 1 was not covered');
            
            zerosMatch = theZeros & ~X(i,:);
            assert(all(theZeros == zerosMatch), 'A 0 was not covered');
        end
    end    
    
    % Triple check: same as above, but no bitwise tests
    for k = 1:nFinalClusters
        os = finalClusters(k).ones;
        zs = ~(finalClusters(k).ones | finalClusters(k).stars);
        is = finalClusters(k).inds;
        
        for d = find(os)
            assert(all(X(is,d) == 1), 'A 1 was not covered');
        end
        
        for d = find(zs)
            assert(all(X(is,d) == 0), 'A 0 was not covered');
        end
    end
    
    % Copy star pattern of the cluster back to each member
    
    S = zeros(size(X));
    IDX(N) = 0;
    for k = 1:nFinalClusters        
        S(finalClusters(k).inds,:) = repmat(finalClusters(k).stars, finalClusters(k).size, 1);
        IDX(finalClusters(k).inds) = k;
    end
    
    % Adjacency matrix
    % Stupid way... hopefully it doesn't dominate runtime
    G = false(N, N);
    for n = 1:N
        neighbors = finalClusters(IDX(n)).inds;
        neighbors(n) = 0;
        G(n,:) = neighbors;
    end
    

end

function d = clusterDist(c1, c2m)
% clusterDist  Distance between two clusters; minimum stars to bring their 
%   d = clusterDist(c1, c2) takes a single cluster in c1 and a column
%   vector in c2
    
    newNStars = arrayfun(@(c2) sum(c1.stars | c2.stars | (c1.ones ~= c2.ones)), c2m)';    
    newSize   = c1.size + [c2m.size];
    d         = newSize .* newNStars - c1.sizeTimesStars - [c2m.sizeTimesStars];
end

function cc = mergeClusters(c1, c2)
% Merge c1 and c2 into cc
    cc.ones  = c1.ones & c2.ones;
    cc.stars = c1.stars | c2.stars | (c1.ones ~= c2.ones);
    assert(sum(c1.inds & c2.inds) == 0, 'Indices must be disjoint!');    
    cc.inds  = c1.inds | c2.inds;
    
    cc.size  = sum(cc.inds);
    cc.nStars = sum(cc.stars);        
    cc.sizeTimesStars = cc.size * cc.nStars;
end



